@@ -138,6 +138,8 @@ STATIC char* txtURL(const char* string, char* outbuff) {
 	return outbuff;
 }
 
+#define PROMOTE(what,from,to) ((what) = ((what) == (from)) ? (to) : (what))
+
 // flags used to mark struct mail so that at the end of the scanning 
 // we output commands lookig that flag
 enum sight {
@@ -188,6 +190,7 @@ STATIC int only_mkdirp;
 STATIC int only_mkfifo;
 STATIC int n_excludes;
 STATIC char **excludes;
+STATIC int no_delete;
 
 // ============================ helpers =====================================
 
@@ -431,13 +434,14 @@ STATIC void save_db(const char* dbname, time_t timestamp){
 STATIC void load_db(const char* dbname){
 	FILE* fd;
 	int fields;
+	int line=0;
 	char new_dbname[PATH_MAX];
 
 	snprintf(new_dbname,PATH_MAX,"%s.mtime",dbname);
 
 	fd = fopen(new_dbname,"r");
 	if (fd == NULL){
-		WARNING(fopen,"unable to load db file '%s'\n",new_dbname);
+		WARNING(fopen,"unable to open db file '%s'\n",new_dbname);
 		lastcheck = 0L;
 	} else {
 		fields = fscanf(fd,"%1$lu",&lastcheck);
@@ -462,6 +466,7 @@ STATIC void load_db(const char* dbname){
 		fields = fscanf(fd,
 			"%1$40s %2$40s %3$" tostring(MAX_EMAIL_NAME_LEN) "[^\n]\n",
 			tmpbuff_1, tmpbuff_2, next_name());
+		line++;
 
 		if (fields == EOF) {
 			// deallocate mail entry
@@ -471,7 +476,8 @@ STATIC void load_db(const char* dbname){
 		
 		// sanity checks
 		if (fields != 3)
-			ERROR(fscanf,"malformed db file '%s', please remove it\n",dbname);
+			ERROR(fscanf, "%s: malformed line %d: %d != 3 fields."
+				" Please remove this db file.\n", dbname, line, fields);
 
 		shatxt(tmpbuff_1, mail(m)->hsha);
 		shatxt(tmpbuff_2, mail(m)->bsha);
@@ -483,7 +489,8 @@ STATIC void load_db(const char* dbname){
 		mail(m)->seen=NOT_SEEN;
 
 		// store it in the hash tables
-		g_hash_table_insert(bsha2mail,GPTR(m),GPTR(m));
+		g_hash_table_insert(bsha2mail,GPTR(m),
+			g_slist_prepend(g_hash_table_lookup(bsha2mail,GPTR(m)),GPTR(m)));
 		g_hash_table_insert(filename2mail,GPTR(m),GPTR(m));
 		
 	} 
@@ -533,9 +540,10 @@ STATIC void analyze_file(const char* dir,const char* file) {
 	unsigned char *addr,*next;
 	int fd;
 	struct stat sb;
-	mail_t alias, bodyalias, m;
+	mail_t alias, m;
 	GChecksum* ctx;
 	gsize ctx_len;
+	GSList *bodyaliases = NULL;
 
 	m = alloc_mail();
 	snprintf(next_name(), MAX_EMAIL_NAME_LEN,"%s/%s",dir,file);
@@ -618,18 +626,25 @@ STATIC void analyze_file(const char* dir,const char* file) {
 		}
 	}
 
-	bodyalias = MAIL(g_hash_table_lookup(bsha2mail,GPTR(m)));
+	bodyaliases = g_hash_table_lookup(bsha2mail,GPTR(m));
 
-	if (bodyalias != 0) {
-		if (hsha_equal(GPTR(bodyalias), GPTR(m))) {
-			COMMAND_COPY(bodyalias,m);
-			mail(m)->seen=SEEN;
-			return;
-		} else {
-			COMMAND_COPYBODY(bodyalias,m);
-			mail(m)->seen=SEEN;
-			return;
+	// some messages with the same body are there
+	if (bodyaliases != NULL) {
+		mail_t firstalias = MAIL(bodyaliases->data);
+		for(; bodyaliases != NULL; bodyaliases = g_slist_next(bodyaliases)) {
+			mail_t bodyalias = MAIL(bodyaliases->data);
+			if (hsha_equal(GPTR(bodyalias), GPTR(m))) {
+				// this one has the same header too
+				COMMAND_COPY(bodyalias,m);
+				PROMOTE(mail(bodyalias)->seen, NOT_SEEN, MOVED);
+				mail(m)->seen=SEEN;
+				return;
+			}
 		}
+		// no full alias, we just recycle the body
+		COMMAND_COPYBODY(firstalias,m);
+		mail(m)->seen=SEEN;
+		return;
 	}
 
 	// we should add that file
@@ -727,7 +742,11 @@ STATIC void generate_deletions(){
 	size_t m;
 
 	for(m=1; m < mailno; m++){
-		if (mail(m)->seen == NOT_SEEN) 
+		if (!no_delete && (mail(m)->seen == NOT_SEEN || mail(m)->seen == MOVED))
+			// normally moved or removed mails are deleted
+			COMMAND_DELETE(m);
+		else if (no_delete && mail(m)->seen == MOVED)
+			// if --no-delete only moved mails should be deleted
 			COMMAND_DELETE(m);
 		else 
 			VERBOSE(seen,"STATUS OF %s %s %s IS %s\n",
@@ -808,7 +827,14 @@ STATIC void extra_sha1sum_file(const char* file) {
 	if (fstat(fd, &sb) == -1) ERROR(fstat,"unable to stat file '%s'\n",file);
 
 	addr = mmap(NULL, sb.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
-	if (addr == MAP_FAILED) ERROR(mmap, "unable to load '%s'\n",file);
+	if (addr == MAP_FAILED){
+		if (sb.st_size == 0) 
+			// empty file
+			;
+		else 
+			// mmap failed
+			ERROR(mmap, "unable to load '%s'\n",file);
+	}
 
 	// calculate sha1
 	fprintf(stdout, "%s  %s\n", 
@@ -816,7 +842,7 @@ STATIC void extra_sha1sum_file(const char* file) {
 		file);
 	g_free(sha1);
 	
-	munmap(addr, sb.st_size);
+	if (addr != MAP_FAILED) munmap(addr, sb.st_size);
 	close(fd);
 }
 
@@ -841,6 +867,7 @@ STATIC struct option long_options[] = {
 	{"symlink"   , no_argument      , NULL, 's'},
 	{"verbose"   , no_argument      , NULL, 'v'},
 	{"dry-run"   , no_argument      , NULL, 'd'},
+	{"no-delete" , no_argument      , NULL, 'n'},
 	{"help"      , no_argument      , NULL, 'h'},
 	{NULL        , no_argument      , NULL, 0},
 };
@@ -866,6 +893,7 @@ STATIC const char* long_options_doc[] = {
 			"Symbolic Link generation mode (short -s)",
 			"Increase program verbosity (printed on stderr, short -v)",
 			"Do not generate a new db file (short -d)",
+			"Do not track deletions (short -n)",
 			"This help screen",
 			NULL
 };
@@ -919,7 +947,7 @@ int main(int argc, char *argv[]) {
 	g_assert(MAIL(GPTR(1)) == 1);
 
 	for(;;) {
-		c = getopt_long(argc, argv, "vhdls", long_options, &option_index);
+		c = getopt_long(argc, argv, "vhndls", long_options, &option_index);
 		if (c == -1) break; // no more args
 		switch (c) {
 			case OPT_MAX_MAILNO:
@@ -948,6 +976,9 @@ int main(int argc, char *argv[]) {
 			case 'd':
 				dry_run = 1;
 			break;
+			case 'n':
+				no_delete = 1;
+			break;
 			case 'l':
 				only_list_subfolders = 1;
 			break;
