@@ -29,7 +29,7 @@ echo "Tested $PATHS paths"
 echo
 
 grep TRACE $ROOT/[0-9]*/log.client* \
-	| sort -u | sed 's?^.*/\([^/]*\):?\1?' | sed 's/TRACE://'
+	| sort -u | sed 's?^.*/\([^/]*\):?\1:?' | sed 's/TRACE://'
 
 echo
 echo "Surely missing leaves (there may be more paths for the same leaf):"
